@@ -126,14 +126,18 @@ class MPRester:
 
         try:
             self.contribs = Client(api_key)
-        except Exception:
+        except Exception as error:
             self.contribs = None
+            warnings.warn(f"Problem loading MPContribs client: {error}")
 
         self._all_resters = []
 
         if notify_db_version:
             raise NotImplementedError("This has not yet been implemented.")
 
+        if not self.endpoint.endswith("/"):
+            self.endpoint += "/"
+
         for cls in BaseRester.__subclasses__():
 
             rester = cls(
@@ -182,7 +186,7 @@ class MPRester:
                 task for task, calc_type in tasks.items() if calc_type in calc_types
             ]
         else:
-            return list(tasks.values())
+            return list(tasks.keys())
 
     def get_structure_by_material_id(
         self, material_id: str, final: bool = True, conventional_unit_cell: bool = False
@@ -233,7 +237,7 @@ class MPRester:
 
         Returns: database version as a string
         """
-        return get(url=self.endpoint + "/heartbeat").json()["db_version"]
+        return get(url=self.endpoint + "heartbeat").json()["db_version"]
 
     def get_materials_id_from_task_id(self, task_id: str) -> Union[str, None]:
         """
